#include "hash.hh"
#include "expr.hh"
#include "stmt.hh"

typedef autoptr<struct TypeInfo> TypeP;
typedef autoptr<struct UserType> UtypeP;
struct TypeInfo: public ptrable
{
    enum
    {
        Integer, Long, Single, Double, String,
        UserType
    } Type;
    
    bool dynamic;            //if resizable (Array, String).
    unsigned fixlength;      //if non-dynamic string.
    UtypeP members;          //if UserType
    
    // If array:
    std::vector<std::pair<unsigned, unsigned> > dimensions;
    // ^first  = beginning index
    // ^second = dimension (size)
    //   example: 1 TO 5   is (1,5)
    //            0 TO 4   is (0,5)
    
    TypeInfo()
    : Type(Integer),dynamic(false),fixlength(0),members(),dimensions()
    { }
};

struct UserType: public std::vector<std::pair<std::string, TypeP> >, public ptrable
{
    hash_set<std::string> Names;
    
    bool DeclareMember(const std::string& name, TypeP p)
    {
        if(Names.find(name) != Names.end()) return false;
        Names.insert(name);
        (*this).push_back(std::make_pair(name, p));
        return true;
    }
};

struct Constant
{
    QBtoken Type; //integer,single,double,long or string.
    exprp value;  
};

struct Subroutine // SUB, FUNCTION, DEF
{
    stmtp code;
};

struct TypeFlagSet
{
    bool is_subroutine:1;
    bool is_scaler    :1;
    bool is_const     :1;
    bool is_array     :1;
    
    // In case of subroutines, they can be DECLARED and later defined.
    bool has_body     :1;
    
    // Types and labels not included here; they are in different namespace.
};

typedef autoptr<struct VarContextDef> ctxp;
struct VarContextDef
{
    TypeP type;
    bool is_static; // STATIC to the given function
    bool is_shared; // SHARED to the subfunctions (not from above)
};

struct SubRoutine
{
    TypeP returnvalue; // 0=SUB, otherwise FUNCTION/DEF
};

struct Scope
{
    /*****
      Method of lookup:
        Check constants. Ignore suffix.
          Match? Fail if had wrong suffix. Otherwise, accept.
      
        Check UTYPE.
          Match? Fail if had a suffix. Otherwise, accept.
        Had suffix? Check by suffix.
        Otherwise, check by DEFtype.
    *******/
    
    hash_map<std::string, UtypeP> types;
    hash_map<std::string, Constant> constants;
    
    // SUBS, FUNCTIONS, DEFs
    hash_map<std::string, SubRoutine> subroutines;
    
    hash_map<std::string, ctxp> vars_UTYPE;
    hash_map<std::string, ctxp> vars_INT;
    hash_map<std::string, ctxp> vars_SINGLE;
    hash_map<std::string, ctxp> vars_DOUBLE;
    hash_map<std::string, ctxp> vars_LONG;
    hash_map<std::string, ctxp> vars_STRING;
    
    hash_map<std::string, ctxp> arrays_UTYPE;
    hash_map<std::string, ctxp> arrays_INT;
    hash_map<std::string, ctxp> arrays_SINGLE;
    hash_map<std::string, ctxp> arrays_DOUBLE;
    hash_map<std::string, ctxp> arrays_LONG;
    hash_map<std::string, ctxp> arrays_STRING;

    hash_map<std::string, TypeFlagSet> usage;
    
    // Labels...
    hash_set<std::string> labels;
};
